@@ -45,6 +45,7 @@
 #include <libgen.h>
 #include <sys/mman.h>
 #include <sys/wait.h>
+#include <stdbool.h>
 #ifdef HAVE_SYS_FSUID_H
 #include <sys/fsuid.h>
 #endif /* HAVE_SYS_FSUID_H */
@@ -83,21 +84,30 @@
 #define MTAB_OPTIONS_LEN 220
 
 /*
- * Maximum length of "share" portion of a UNC. I have no idea if this is at
- * all valid. According to MSDN, the typical max length of any component is
- * 255, so use that here.
+ * Max share name, username, password and domain sizes match the kernel's
+ * allowances for these string sizes which in turn match Microsoft's
+ * documentation.
  */
+
+/* Max length of the share name portion of a UNC. Share names over 80
+ * characters cannot be accessed via commandline in Windows 2000/XP. */
 #define MAX_SHARE_LEN 256
 
-/* max length of username (somewhat made up here) */
-#define MAX_USERNAME_SIZE 32
+/* Max user name length. */
+#define MAX_USERNAME_SIZE 256
+
+/* Max domain size. */
+#define MAX_DOMAIN_SIZE 256
+
+/* Max password size. */
+#define MOUNT_PASSWD_SIZE 512
+
+
 
 #ifndef SAFE_FREE
 #define SAFE_FREE(x) do { if ((x) != NULL) {free(x); x = NULL; } } while (0)
 #endif
 
-#define MOUNT_PASSWD_SIZE 128
-#define MAX_DOMAIN_SIZE 64
 
 /*
  * mount.cifs has been the subject of many "security" bugs that have arisen
@@ -158,6 +168,7 @@
 #define OPT_CRUID      29
 #define OPT_BKUPUID    30
 #define OPT_BKUPGID    31
+#define OPT_NOFAIL     32
 
 #define MNT_TMP_FILE "/.mtab.cifs.XXXXXX"
 
@@ -177,10 +188,11 @@ struct parsed_mount_info {
 	unsigned int fakemnt:1;
 	unsigned int nomtab:1;
 	unsigned int verboseflag:1;
+	unsigned int nofail:1;
 };
 
-const char *thisprogram;
-const char *cifs_fstype = "cifs";
+static const char *thisprogram;
+static const char *cifs_fstype = "cifs";
 
 static int parse_unc(const char *unc_name, struct parsed_mount_info *parsed_info);
 
@@ -313,52 +325,6 @@ static int set_password(struct parsed_mount_info *parsed_info, const char *src)
 	return 0;
 }
 
-/*
- * Parse a username string into parsed_mount_info fields. The format is:
- *
- * DOMAIN\username%password
- *
- * ...obviously the only required component is "username". The source string
- * is modified in the process, but it should remain unchanged at the end.
- */
-static int parse_username(char *rawuser, struct parsed_mount_info *parsed_info)
-{
-	char *user, *password, slash;
-	int rc = 0;
-
-	/* everything after first % sign is a password */
-	password = strchr(rawuser, '%');
-	if (password) {
-		rc = set_password(parsed_info, password + 1);
-		if (rc)
-			return rc;
-		*password = '\0';
-	}
-
-	/* everything after first '/' or '\' is a username */
-	user = strchr(rawuser, '/');
-	if (!user)
-		user = strchr(rawuser, '\\');
-
-	/* everything before that slash is a domain */
-	if (user) {
-		slash = *user;
-		*user = '\0';
-		strlcpy(parsed_info->domain, rawuser,
-			sizeof(parsed_info->domain));
-		*(user++) = slash;
-	} else {
-		user = rawuser;
-	}
-
-	strlcpy(parsed_info->username, user, sizeof(parsed_info->username));
-	parsed_info->got_user = 1;
-	if (password)
-		*password = '%';
-
-	return 0;
-}
-
 #ifdef HAVE_LIBCAP_NG
 static int
 drop_capabilities(int parent)
@@ -495,7 +461,7 @@ free_caps:
 static int
 toggle_dac_capability(int writable, int enable)
 {
-	int rc;
+	int rc = 0;
 	cap_t caps;
 	cap_value_t capability = writable ? CAP_DAC_OVERRIDE : CAP_DAC_READ_SEARCH;
 
@@ -590,6 +556,7 @@ static int open_cred_file(char *file_name,
 	i = access(file_name, R_OK);
 	if (i) {
 		toggle_dac_capability(0, 0);
+		i = errno;
 		goto return_i;
 	}
 
@@ -622,9 +589,9 @@ static int open_cred_file(char *file_name,
 		/* parse next token */
 		switch (parse_cred_line(line_buf + i, &temp_val)) {
 		case CRED_USER:
-			i = parse_username(temp_val, parsed_info);
-			if (i)
-				goto return_i;
+			strlcpy(parsed_info->username, temp_val,
+				sizeof(parsed_info->username));
+			parsed_info->got_user = 1;
 			break;
 		case CRED_PASS:
 			i = set_password(parsed_info, temp_val);
@@ -787,6 +754,8 @@ static int parse_opt_token(const char *token)
 		return OPT_BKUPUID;
 	if (strncmp(token, "backupgid", 9) == 0)
 		return OPT_BKUPGID;
+	if (strncmp(token, "nofail", 6) == 0)
+		return OPT_NOFAIL;
 
 	return OPT_ERROR;
 }
@@ -871,20 +840,9 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 					return EX_USAGE;
 				}
 			} else {
-				/* domain/username%password  + NULL term. */
-				const size_t max = MAX_DOMAIN_SIZE +
-						   MAX_USERNAME_SIZE +
-						   MOUNT_PASSWD_SIZE + 2 + 1;
-				if (strnlen(value, max) >= max) {
-					fprintf(stderr, "username too long\n");
-					return EX_USAGE;
-				}
-				rc = parse_username(value, parsed_info);
-				if (rc) {
-					fprintf(stderr,
-						"problem parsing username\n");
-					return rc;
-				}
+				strlcpy(parsed_info->username, value,
+					sizeof(parsed_info->username));
+				parsed_info->got_user = 1;
 				goto nocopy;
 			}
 
@@ -977,58 +935,56 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 				goto nocopy;
 
 			got_uid = 1;
+			pw = getpwnam(value);
+			if (pw) {
+				uid = pw->pw_uid;
+				goto nocopy;
+			}
+
 			errno = 0;
 			uid = strtoul(value, &ep, 10);
 			if (errno == 0 && *ep == '\0')
 				goto nocopy;
 
-			pw = getpwnam(value);
-			if (pw == NULL) {
-				fprintf(stderr, "bad user name \"%s\"\n", value);
-				return EX_USAGE;
-			}
-
-			uid = pw->pw_uid;
-			goto nocopy;
-
+			fprintf(stderr, "bad option uid=\"%s\"\n", value);
+			return EX_USAGE;
 		case OPT_CRUID:
 			if (!value || !*value)
 				goto nocopy;
 
 			got_cruid = 1;
+			pw = getpwnam(value);
+			if (pw) {
+				cruid = pw->pw_uid;
+				goto nocopy;
+			}
+
 			errno = 0;
 			cruid = strtoul(value, &ep, 10);
 			if (errno == 0 && *ep == '\0')
 				goto nocopy;
 
-			pw = getpwnam(value);
-			if (pw == NULL) {
-				fprintf(stderr, "bad user name \"%s\"\n", value);
-				return EX_USAGE;
-			}
-			cruid = pw->pw_uid;
-			goto nocopy;
-
+			fprintf(stderr, "bad option: cruid=\"%s\"\n", value);
+			return EX_USAGE;
 		case OPT_GID:
 			if (!value || !*value)
 				goto nocopy;
 
 			got_gid = 1;
+			gr = getgrnam(value);
+			if (gr) {
+				gid = gr->gr_gid;
+				goto nocopy;
+			}
+
 			errno = 0;
 			gid = strtoul(value, &ep, 10);
 			if (errno == 0 && *ep == '\0')
 				goto nocopy;
 
-			gr = getgrnam(value);
-			if (gr == NULL) {
-				fprintf(stderr, "bad group name \"%s\"\n", value);
-				return EX_USAGE;
-			}
-
-			gid = gr->gr_gid;
-			goto nocopy;
-
-		/* fmask fall through to file_mode */
+			fprintf(stderr, "bad option: gid=\"%s\"\n", value);
+			return EX_USAGE;
+		/* fmask falls through to file_mode */
 		case OPT_FMASK:
 			fprintf(stderr,
 				"WARNING: CIFS mount option 'fmask' is\
@@ -1067,23 +1023,15 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 					"WARNING: '%s' not expressed in octal.\n",
 					data);
 			break;
-
-		/* the following mount options should be
-		   stripped out from what is passed into the kernel
-		   since these options are best passed as the
-		   mount flags rather than redundantly to the kernel
-		   and could generate spurious warnings depending on the
-		   level of the corresponding cifs vfs kernel code */
 		case OPT_NO_SUID:
 			*filesys_flags |= MS_NOSUID;
-			break;
+			goto nocopy;
 		case OPT_SUID:
 			*filesys_flags &= ~MS_NOSUID;
-			break;
+			goto nocopy;
 		case OPT_NO_DEV:
 			*filesys_flags |= MS_NODEV;
-			break;
-		/* nolock || nobrl */
+			goto nocopy;
 		case OPT_NO_LOCK:
 			*filesys_flags &= ~MS_MANDLOCK;
 			break;
@@ -1095,17 +1043,17 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 			goto nocopy;
 		case OPT_DEV:
 			*filesys_flags &= ~MS_NODEV;
-			break;
+			goto nocopy;
 		case OPT_NO_EXEC:
 			*filesys_flags |= MS_NOEXEC;
-			break;
+			goto nocopy;
 		case OPT_EXEC:
 			*filesys_flags &= ~MS_NOEXEC;
-			break;
+			goto nocopy;
 		case OPT_GUEST:
 			parsed_info->got_user = 1;
 			parsed_info->got_password = 1;
-			break;
+			goto nocopy;
 		case OPT_RO:
 			*filesys_flags |= MS_RDONLY;
 			goto nocopy;
@@ -1114,7 +1062,7 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 			goto nocopy;
 		case OPT_REMOUNT:
 			*filesys_flags |= MS_REMOUNT;
-			break;
+			goto nocopy;
 		case OPT_IGNORE:
 			goto nocopy;
 		case OPT_BKUPUID:
@@ -1155,6 +1103,9 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 
 			bkupgid = gr->gr_gid;
 			goto nocopy;
+		case OPT_NOFAIL:
+			parsed_info->nofail = 1;
+			goto nocopy;
 		}
 
 		/* check size before copying option to buffer */
@@ -1182,6 +1133,7 @@ nocopy:
 		data = next_keyword;
 	}
 
+
 	/* special-case the uid and gid */
 	if (got_uid) {
 		word_len = snprintf(txtbuf, sizeof(txtbuf), "%u", uid);
@@ -1202,7 +1154,7 @@ nocopy:
 	if (got_cruid) {
 		word_len = snprintf(txtbuf, sizeof(txtbuf), "%u", cruid);
 
-		/* comma + "cruid=" + terminating NULL == 6 */
+		/* comma + "cruid=" + terminating NULL == 8 */
 		if (out_len + word_len + 8 > MAX_OPTIONS_LEN) {
 			fprintf(stderr, "Options string too long\n");
 			return EX_USAGE;
@@ -1289,33 +1241,19 @@ static int parse_unc(const char *unc_name, struct parsed_mount_info *parsed_info
 		return EX_USAGE;
 	}
 
-	/* Set up "host" and "share" pointers based on UNC format. */
 	if (strncmp(unc_name, "//", 2) && strncmp(unc_name, "\\\\", 2)) {
-		/*
-		 * check for nfs syntax (server:/share/prepath)
-		 *
-		 * FIXME: IPv6 addresses?
-		 */
-		host = unc_name;
-		share = strchr(host, ':');
-		if (!share) {
-			fprintf(stderr, "mount.cifs: bad UNC (%s)\n", unc_name);
-			return EX_USAGE;
-		}
-		hostlen = share - host;
-		share++;
-		if (*share == '/')
-			++share;
-	} else {
-		host = unc_name + 2;
-		hostlen = strcspn(host, "/\\");
-		if (!hostlen) {
-			fprintf(stderr, "mount.cifs: bad UNC (%s)\n", unc_name);
-			return EX_USAGE;
-		}
-		share = host + hostlen + 1;
+		fprintf(stderr, "mount.cifs: bad UNC (%s)\n", unc_name);
+		return EX_USAGE;
 	}
 
+	host = unc_name + 2;
+	hostlen = strcspn(host, "/\\");
+	if (!hostlen) {
+		fprintf(stderr, "mount.cifs: bad UNC (%s)\n", unc_name);
+		return EX_USAGE;
+	}
+	share = host + hostlen + 1;
+
 	if (hostlen + 1 > sizeof(parsed_info->host)) {
 		fprintf(stderr, "mount.cifs: host portion of UNC too long\n");
 		return EX_USAGE;
@@ -1570,23 +1508,29 @@ add_mtab_exit:
 static int
 del_mtab(char *mountpoint)
 {
-	int tmprc, rc = 0;
+	int len, tmprc, rc = 0;
 	FILE *mnttmp, *mntmtab;
 	struct mntent *mountent;
-	char *mtabfile, *mtabdir, *mtabtmpfile;
+	char *mtabfile, *mtabdir, *mtabtmpfile = NULL;
 
 	mtabfile = strdup(MOUNTED);
-	mtabdir = dirname(mtabfile);
-	mtabdir = realloc(mtabdir, strlen(mtabdir) + strlen(MNT_TMP_FILE) + 2);
-	if (!mtabdir) {
-		fprintf(stderr, "del_mtab: cannot determine current mtab path");
+	if (!mtabfile) {
+		fprintf(stderr, "del_mtab: cannot strdup MOUNTED\n");
 		rc = EX_FILEIO;
 		goto del_mtab_exit;
 	}
 
-	mtabtmpfile = strcat(mtabdir, MNT_TMP_FILE);
+	mtabdir = dirname(mtabfile);
+	len = strlen(mtabdir) + strlen(MNT_TMP_FILE);
+	mtabtmpfile = malloc(len + 1);
 	if (!mtabtmpfile) {
-		fprintf(stderr, "del_mtab: cannot allocate memory to tmp file");
+		fprintf(stderr, "del_mtab: cannot allocate memory to tmp file\n");
+		rc = EX_FILEIO;
+		goto del_mtab_exit;
+	}
+
+	if (sprintf(mtabtmpfile, "%s%s", mtabdir, MNT_TMP_FILE) != len) {
+		fprintf(stderr, "del_mtab: error writing new string\n");
 		rc = EX_FILEIO;
 		goto del_mtab_exit;
 	}
@@ -1594,14 +1538,14 @@ del_mtab(char *mountpoint)
 	atexit(unlock_mtab);
 	rc = lock_mtab();
 	if (rc) {
-		fprintf(stderr, "del_mtab: cannot lock mtab");
+		fprintf(stderr, "del_mtab: cannot lock mtab\n");
 		rc = EX_FILEIO;
 		goto del_mtab_exit;
 	}
 
 	mtabtmpfile = mktemp(mtabtmpfile);
 	if (!mtabtmpfile) {
-		fprintf(stderr, "del_mtab: cannot setup tmp file destination");
+		fprintf(stderr, "del_mtab: cannot setup tmp file destination\n");
 		rc = EX_FILEIO;
 		goto del_mtab_exit;
 	}
@@ -1649,7 +1593,8 @@ del_mtab(char *mountpoint)
 
 del_mtab_exit:
 	unlock_mtab();
-	free(mtabdir);
+	free(mtabtmpfile);
+	free(mtabfile);
 	return rc;
 
 del_mtab_error:
@@ -1687,6 +1632,66 @@ drop_child_privs(void)
 	return 0;
 }
 
+/*
+ * If systemd is running and systemd-ask-password --
+ * is available, then use that else fallback on getpass(..)
+ *
+ * Returns: @input or NULL on error
+ */
+static char*
+get_password(const char *prompt, char *input, int capacity)
+{
+#ifdef ENABLE_SYSTEMD
+	int is_systemd_running;
+	struct stat a, b;
+
+	/* We simply test whether the systemd cgroup hierarchy is
+	 * mounted */
+	is_systemd_running = (lstat("/sys/fs/cgroup", &a) == 0)
+		&& (lstat("/sys/fs/cgroup/systemd", &b) == 0)
+		&& (a.st_dev != b.st_dev);
+
+	if (is_systemd_running) {
+		char *cmd, *ret;
+		FILE *ask_pass_fp = NULL;
+
+		cmd = ret = NULL;
+		if (asprintf(&cmd, "systemd-ask-password \"%s\"", prompt) >= 0) {
+			ask_pass_fp = popen (cmd, "re");
+			free (cmd);
+		}
+
+		if (ask_pass_fp) {
+			ret = fgets(input, capacity, ask_pass_fp);
+			pclose(ask_pass_fp);
+		}
+
+		if (ret) {
+			int len = strlen(input);
+			if (input[len - 1] == '\n')
+				input[len - 1] = '\0';
+			return input;
+		}
+	}
+#endif
+
+	/*
+	 * Falling back to getpass(..)
+	 * getpass is obsolete, but there's apparently nothing that replaces it
+	 */
+	char *tmp_pass = getpass(prompt);
+	if (!tmp_pass)
+		return NULL;
+
+	strncpy(input, tmp_pass, capacity - 1);
+	input[capacity - 1] = '\0';
+
+	/* zero-out the static buffer */
+	memset(tmp_pass, 0, strlen(tmp_pass));
+
+	return input;
+}
+
 static int
 assemble_mountinfo(struct parsed_mount_info *parsed_info,
 		   const char *thisprogram, const char *mountpoint,
@@ -1768,14 +1773,20 @@ assemble_mountinfo(struct parsed_mount_info *parsed_info,
 	}
 
 	if (!parsed_info->got_password) {
-		/* getpass is obsolete, but there's apparently nothing that replaces it */
-		char *tmp_pass = getpass("Password: ");
-		if (!tmp_pass) {
+		char tmp_pass[MOUNT_PASSWD_SIZE + 1];
+		char *prompt = NULL;
+
+		if(asprintf(&prompt, "Password for %s@%s: ", parsed_info->username, orig_dev) < 0)
+			prompt = NULL;
+
+		if (get_password(prompt ? prompt : "Password: ", tmp_pass, MOUNT_PASSWD_SIZE + 1)) {
+			rc = set_password(parsed_info, tmp_pass);
+		} else {
 			fprintf(stderr, "Error reading password, exiting\n");
 			rc = EX_SYSERR;
-			goto assemble_exit;
 		}
-		rc = set_password(parsed_info, tmp_pass);
+
+		free(prompt);
 		if (rc)
 			goto assemble_exit;
 	}
@@ -1860,8 +1871,8 @@ restore_privs:
 		if (dacrc)
 			rc = rc ? rc : dacrc;
 	} else {
-		setfsuid(oldfsuid);
-		setfsgid(oldfsgid);
+		uid_t __attribute__((unused)) uignore = setfsuid(oldfsuid);
+		gid_t __attribute__((unused)) gignore = setfsgid(oldfsgid);
 	}
 
 	return rc;
@@ -1904,7 +1915,7 @@ int main(int argc, char **argv)
 		thisprogram = "mount.cifs";
 
 	/* allocate parsed_info as shared anonymous memory range */
-	parsed_info = mmap(0, sizeof(*parsed_info), PROT_READ | PROT_WRITE,
+	parsed_info = mmap((void *)0, sizeof(*parsed_info), PROT_READ | PROT_WRITE,
 			   MAP_ANONYMOUS | MAP_SHARED, -1, 0);
 	if (parsed_info == (struct parsed_mount_info *) -1) {
 		parsed_info = NULL;
@@ -1956,7 +1967,7 @@ int main(int argc, char **argv)
 		}
 	}
 
-	if (argc < 3 || argv[optind] == NULL || argv[optind + 1] == NULL) {
+	if (argc < optind + 2) {
 		rc = mount_usage(stderr);
 		goto mount_exit;
 	}
@@ -2015,7 +2026,7 @@ int main(int argc, char **argv)
 mount_retry:
 	if (!currentaddress) {
 		fprintf(stderr, "Unable to find suitable address.\n");
-		rc = EX_SYSERR;
+		rc = parsed_info->nofail ? 0 : EX_FAIL;
 		goto mount_exit;
 	}
 	strlcpy(options, "ip=", options_size);
@@ -2086,7 +2097,8 @@ mount_retry:
 			if (!already_uppercased &&
 			    uppercase_string(parsed_info->host) &&
 			    uppercase_string(parsed_info->share) &&
-			    uppercase_string(parsed_info->prefix)) {
+			    uppercase_string(parsed_info->prefix) &&
+			    uppercase_string(orig_dev)) {
 				fprintf(stderr,
 					"Retrying with upper case share name\n");
 				already_uppercased = 1;
